"""
    spout
    ====

    A simple framework that makes it easy to work with data streams in Python.

    :copyright: (c) 2014 by Jamie Davies.
    :license: MIT, see LICENSE for more details.
"""

__version__ = '0.1.5'
__author__ = 'Jamie Davies'



from streams import BufferedStream, Stream
from multiprocessing import Process
from twython import TwythonStreamer
import json
import sys


class FileInputStream(Stream):
    """
    Loads the lines of a text file into a Stream for processing.
    """

    def __init__(self, filename):
        self.input = open(filename, 'r')
        self.next_line = None

    def has_next(self):
        if self.next_line is not None:
            return True

        try:
            self.next_line = self.input.next()
        except StopIteration:
            self.next_line = None

        return self.next_line is not None

    def next(self):
        if not self.has_next():
            raise Exception("No more elements in file!")

        to_return = self.next_line
        self.next_line = None

        return to_return

    def __del__(self):
        self.input.close()


class JSONInputStream(FileInputStream):
    """
    Loads a file containing a separate JSON object on each line into a Stream for processing.
    """

    def __init__(self, filename):
        super(JSONInputStream, self).__init__(filename)

    def next(self):
        return json.loads(super(JSONInputStream, self).next())

    def __del__(self):
        super(JSONInputStream, self).__del__()


class TweetStream(BufferedStream):
    """
    Publishes tweets fetched from the Twitter Sample stream to a BufferedStream for processing.
    """
    def __init__(self, buf, app_key, app_secret, access_token, access_token_secret):
        super(TweetStream, self).__init__(buf)
        self.source = self.TweetSource(self, app_key, app_secret, access_token,
                                       access_token_secret)
        self.worker = Process(target=self.source.statuses.sample)
        self.worker.daemon = True

    def connect(self):
        super(TweetStream, self).connect()
        self.worker.start()

    def disconnect(self):
        super(TweetStream, self).disconnect()

        self.source.disconnect()
        self.worker.terminate()


    class TweetSource(TwythonStreamer):
        """
        Acts as the 'listener' to the official Twitter Sample stream.
        """
        def __init__(self, publish_stream, app_key, app_secret, access_token, access_token_secret):
            super(TweetStream.TweetSource, self).__init__(app_key, app_secret, access_token,
                                                          access_token_secret)
            self.publish_stream = publish_stream

        def on_success(self, data):
            if 'text' in data:
                self.publish_stream.register(data)

        def on_error(self, status_code, data):
            sys.stderr.write(status_code, data)




from setuptools import setup

setup(
    name="spout",
    version="0.1.5",
    author="Jamie Davies",
    author_email="jamie@jamiedavies.me",
    packages=['spout'],
    url="http://github.com/daviesjamie/spout",
    license="MIT",
    description="A simple framework that makes it easy to work with data streams in Python.",
    install_requires=[
        "twython >= 3.1.2"
    ],
    long_description="""
Spout is a small and simple framework that makes it easy to work with data
streams in Python. In particular, Spout was designed with the processing and
consumption of live data sources in mind.


How it works
------------

At the heart of Spout is the concept of a Stream (which is defined in an
abstract `Stream` class). This defines the basic operations that can be
performed upon a data stream:

Mapping
    The items in one stream can be "mapped" to another stream. This is done by
    applying a supplied Function to each item in the input stream, to produce
    another output stream.

    stream.map(Function)

Filtering
    The items in a stream can be "filtered", so that the resultant stream only
    contains items that match a given criteria. This is done by using a supplied
    Predicate to test each item in the input stream, and copies it to the output
    stream if it passes the test criteria.

    stream.filter(Predicate)

Processing (Consuming)
    The items in a stream are used in some calculations or functionality that
    provides no further output to the stream. This is done by applying the supplied
    Operation to each item in the stream.

    stream.for_each(Operation)


Usage
-----

To use Spout, you first need to create an input data stream. A data stream is simply an
instantiation of a Stream or any of its children (which can be found in the
streams.py file). The Stream class has been specifically designed so that it
is easy to extend and wrap around currently existing data sources that you might
have, such as files or databases.

Some existing examples of stream data sources can by found in sources.py.

For example, to create a Stream out of the lines in a plain text file:

    from spout.sources import FileInputStream
    s = FileInputStream("test.txt")

Now that you have your data in a stream, you simply have to process it! This can
be done by creating and using your own Functions, Predicates or Operations
(see above).

For example, to print out all the lines in a text file that start with a digit,
but with the digit stripped, we can create our own Predicate and Function
and pass these to the .filter() and .map() functions:

    from spout.sources import FileInputStream
    from spout.structs import Function, Predicate
    from spout.outputs import PrintOperation


    class StartsWithDigit(Predicate):
        def test(self, obj):
            return obj[0].is_digit()


    class StripFirstChar(Function):
        def apply(self, input):
            return input[1:]


    s = FileInputStream("test.txt")
    s \
        .filter(StartsWithDigit()) \
        .map(StripFirstChar()) \
        .for_each(PrintOperation())


Installation
------------

Spout is available in the Python Package Index (PyPI), and so the easiest way to
install it is through pip:

    $ pip install spout

However, it is also possible to install the repository from the source, through
the setup.py utility:

    $ python setup.py install


Credits
-------

The inspiration for Spout's fluent interface came largely from the OpenIMAJ streaming framework.
http://www.openimaj.org/
""",
)



from structs import Operation
import json


class PrintOperation(Operation):
    """
    Simple operation to print whatever input is supplied.
    """
    def perform(self, obj):
        print obj


class FileOutputOperation(Operation):
    """
    Operation that writes each input item onto a separate line in a file.
    """
    def __init__(self, filename):
        self.output = open(filename, 'w')

    def perform(self, obj):
        self.output.write(obj + '\n')

    def __del__(self):
        self.output.flush()
        self.output.close()


class JSONFileOutputOperation(FileOutputOperation):
    """
    Operation that writes all items it receives as JSON objects on separate lines in a file.
    """
    def __init__(self, filename):
        super(JSONFileOutputOperation, self).__init__(filename)


    def perform(self, obj):
        self.output.write(json.dumps(obj) + '\n')

    def __del__(self):
        super(JSONFileOutputOperation, self).__del__()
